"""

"""
import time

import hiro
from flask import Blueprint

from flask_limiter.constants import ConfigVars


def test_redis_request_slower_than_fixed_window(redis_connection, extension_factory):
    app, limiter = extension_factory(
        {
            ConfigVars.DEFAULT_LIMITS: "5 per second",
            ConfigVars.STORAGE_URI: "redis://localhost:46379",
            ConfigVars.STRATEGY: "fixed-window",
            ConfigVars.HEADERS_ENABLED: True,
        }
    )

    @app.route("/t1")
    def t1():
        time.sleep(1.1)
        return "t1"

    with app.test_client() as cli:
        resp = cli.get("/t1")
        assert resp.headers["X-RateLimit-Remaining"] == "5"


def test_redis_request_slower_than_moving_window(redis_connection, extension_factory):
    app, limiter = extension_factory(
        {
            ConfigVars.DEFAULT_LIMITS: "5 per second",
            ConfigVars.STORAGE_URI: "redis://localhost:46379",
            ConfigVars.STRATEGY: "moving-window",
            ConfigVars.HEADERS_ENABLED: True,
        }
    )

    @app.route("/t1")
    def t1():
        time.sleep(1.1)
        return "t1"

    with app.test_client() as cli:
        resp = cli.get("/t1")
        assert resp.headers["X-RateLimit-Remaining"] == "5"


def test_dynamic_limits(extension_factory):
    app, limiter = extension_factory(
        {ConfigVars.STRATEGY: "moving-window", ConfigVars.HEADERS_ENABLED: True}
    )

    def func(*a):
        return "1/second; 2/minute"

    @app.route("/t1")
    @limiter.limit(func)
    def t1():
        return "t1"

    with hiro.Timeline().freeze() as timeline:
        with app.test_client() as cli:
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 429
            timeline.forward(2)
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 429


def test_invalid_ratelimit_key(extension_factory):
    app, limiter = extension_factory({ConfigVars.HEADERS_ENABLED: True})

    def func(*a):
        return None

    @app.route("/t1")
    @limiter.limit("2/second", key_func=func)
    def t1():
        return "t1"

    with app.test_client() as cli:
        cli.get("/t1")
        cli.get("/t1")
        cli.get("/t1")
        assert cli.get("/t1").status_code == 200
        limiter.limit("1/second", key_func=lambda: "key")(t1)
        cli.get("/t1")
        assert cli.get("/t1").status_code == 429


def test_custom_key_prefix_with_headers(redis_connection, extension_factory):
    app1, limiter1 = extension_factory(
        key_prefix="moo", storage_uri="redis://localhost:46379", headers_enabled=True
    )
    app2, limiter2 = extension_factory(
        key_prefix="cow", storage_uri="redis://localhost:46379", headers_enabled=True
    )

    @app1.route("/test")
    @limiter1.limit("1/minute")
    def t1():
        return "app1 test"

    @app2.route("/test")
    @limiter2.limit("1/minute")
    def t2():
        return "app2 test"

    with app1.test_client() as cli:
        resp = cli.get("/test")
        assert 200 == resp.status_code
        resp = cli.get("/test")
        assert resp.headers.get("Retry-After") == str(60)
        assert 429 == resp.status_code
    with app2.test_client() as cli:
        resp = cli.get("/test")
        assert 200 == resp.status_code
        resp = cli.get("/test")
        assert resp.headers.get("Retry-After") == str(60)
        assert 429 == resp.status_code


def test_default_limits_with_per_route_limit(extension_factory):
    app, limiter = extension_factory(application_limits=["3/minute"])

    @app.route("/explicit")
    @limiter.limit("1/minute")
    def explicit():
        return "explicit"

    @app.route("/default")
    def default():
        return "default"

    with app.test_client() as cli:
        with hiro.Timeline().freeze() as timeline:
            assert 200 == cli.get("/explicit").status_code
            assert 429 == cli.get("/explicit").status_code
            assert 200 == cli.get("/default").status_code
            assert 429 == cli.get("/default").status_code
            timeline.forward(60)
            assert 200 == cli.get("/explicit").status_code
            assert 200 == cli.get("/default").status_code


def test_application_limits_from_config(extension_factory):
    app, limiter = extension_factory(
        config={
            ConfigVars.APPLICATION_LIMITS: "4/second",
            ConfigVars.DEFAULT_LIMITS: "1/second",
            ConfigVars.DEFAULT_LIMITS_PER_METHOD: True,
        }
    )

    @app.route("/root")
    def root():
        return "null"

    @app.route("/test", methods=["GET", "PUT"])
    @limiter.limit("3/second", methods=["GET"])
    def test():
        return "test"

    with app.test_client() as cli:
        with hiro.Timeline() as timeline:
            assert cli.get("/root").status_code == 200
            assert cli.get("/root").status_code == 429
            assert cli.get("/test").status_code == 200
            assert cli.get("/test").status_code == 200
            assert cli.get("/test").status_code == 429
            timeline.forward(1)
            assert cli.get("/test").status_code == 200
            assert cli.get("/test").status_code == 200
            assert cli.get("/test").status_code == 200
            assert cli.get("/test").status_code == 429
            timeline.forward(1)
            assert cli.put("/test").status_code == 200
            assert cli.put("/test").status_code == 429
            assert cli.get("/test").status_code == 200
            assert cli.get("/root").status_code == 200
            assert cli.get("/test").status_code == 429


def test_endpoint_with_dot_but_not_blueprint(extension_factory):
    """
    https://github.com/alisaifee/flask-limiter/issues/336
    """
    app, limiter = extension_factory(default_limits=["2/day"])

    def route():
        return "42"

    app.add_url_rule("/teapot/iam", "_teapot.iam", route)
    bp = Blueprint("teapot", __name__, url_prefix="/teapot")

    @bp.route("/")
    def bp_route():
        return "43"

    app.register_blueprint(bp)
    limiter.limit("1/day")(bp)

    with app.test_client() as cli:
        assert cli.get("/teapot/iam").status_code == 200
        assert cli.get("/teapot/iam").status_code == 200
        assert cli.get("/teapot/iam").status_code == 429
        assert cli.get("/teapot/").status_code == 200
        assert cli.get("/teapot/").status_code == 429



import json
from unittest.mock import patch

import hiro
from flask import make_response

from flask_limiter.constants import ConfigVars


def test_error_message(extension_factory):
    app, limiter = extension_factory({ConfigVars.DEFAULT_LIMITS: "1 per day"})

    @app.route("/")
    def null():
        return ""

    with app.test_client() as cli:

        @app.errorhandler(429)
        def ratelimit_handler(e):
            return make_response(
                '{"error" : "rate limit %s"}' % str(e.description), 429
            )

        cli.get("/")
        assert "1 per 1 day" in cli.get("/").data.decode()

        assert {"error": "rate limit 1 per 1 day"} == json.loads(
            cli.get("/").data.decode()
        )


def test_custom_error_message(extension_factory):
    app, limiter = extension_factory()

    @app.errorhandler(429)
    def ratelimit_handler(e):
        return make_response(e.description, 429)

    def l1():
        return "1/second"

    def e1():
        return "dos"

    @app.route("/t1")
    @limiter.limit("1/second", error_message="uno")
    def t1():
        return "1"

    @app.route("/t2")
    @limiter.limit(l1, error_message=e1)
    def t2():
        return "2"

    s1 = limiter.shared_limit("1/second", scope="error_message", error_message="tres")

    @app.route("/t3")
    @s1
    def t3():
        return "3"

    with hiro.Timeline().freeze():
        with app.test_client() as cli:
            cli.get("/t1")
            resp = cli.get("/t1")
            assert 429 == resp.status_code
            assert resp.data == b"uno"
            cli.get("/t2")
            resp = cli.get("/t2")
            assert 429 == resp.status_code
            assert resp.data == b"dos"
            cli.get("/t3")
            resp = cli.get("/t3")
            assert 429 == resp.status_code
            assert resp.data == b"tres"


def test_swallow_error(extension_factory):
    app, limiter = extension_factory(
        {
            ConfigVars.DEFAULT_LIMITS: "1 per day",
            ConfigVars.HEADERS_ENABLED: True,
            ConfigVars.SWALLOW_ERRORS: True,
        }
    )

    @app.route("/")
    def null():
        return "ok"

    with app.test_client() as cli:
        with patch("limits.strategies.FixedWindowRateLimiter.hit") as hit:

            def raiser(*a, **k):
                raise Exception

            hit.side_effect = raiser
            assert "ok" in cli.get("/").data.decode()
        with patch(
            "limits.strategies.FixedWindowRateLimiter.get_window_stats"
        ) as get_window_stats:

            def raiser(*a, **k):
                raise Exception

            get_window_stats.side_effect = raiser
            assert "ok" in cli.get("/").data.decode()


def test_swallow_error_conditional_deduction(extension_factory):
    def conditional_deduct(_):
        return True

    app, limiter = extension_factory(
        {
            ConfigVars.DEFAULT_LIMITS: "1 per day",
            ConfigVars.SWALLOW_ERRORS: True,
            ConfigVars.DEFAULT_LIMITS_DEDUCT_WHEN: conditional_deduct,
        }
    )

    @app.route("/")
    def null():
        return "ok"

    with app.test_client() as cli:
        with patch("limits.strategies.FixedWindowRateLimiter.hit") as hit:

            def raiser(*a, **k):
                raise Exception

            hit.side_effect = raiser
            assert "ok" in cli.get("/").data.decode()


def test_no_swallow_error(extension_factory):
    app, limiter = extension_factory(
        {ConfigVars.DEFAULT_LIMITS: "1 per day", ConfigVars.HEADERS_ENABLED: True}
    )

    @app.route("/")
    def null():
        return "ok"

    @app.errorhandler(500)
    def e500(e):
        return str(e.original_exception), 500

    def raiser(*a, **k):
        raise Exception("underlying")

    with app.test_client() as cli:
        with patch("limits.strategies.FixedWindowRateLimiter.hit") as hit:
            hit.side_effect = raiser
            assert 500 == cli.get("/").status_code
            assert "underlying" == cli.get("/").data.decode()
        with patch(
            "limits.strategies.FixedWindowRateLimiter.get_window_stats"
        ) as get_window_stats:
            get_window_stats.side_effect = raiser
            assert 500 == cli.get("/").status_code
            assert "underlying" == cli.get("/").data.decode()


def test_no_swallow_error_conditional_deduction(extension_factory):
    def conditional_deduct(_):
        return True

    app, limiter = extension_factory(
        {
            ConfigVars.DEFAULT_LIMITS: "1 per day",
            ConfigVars.SWALLOW_ERRORS: False,
            ConfigVars.DEFAULT_LIMITS_DEDUCT_WHEN: conditional_deduct,
        }
    )

    @app.route("/")
    def null():
        return "ok"

    with app.test_client() as cli:
        with patch("limits.strategies.FixedWindowRateLimiter.hit") as hit:

            def raiser(*a, **k):
                raise Exception

            hit.side_effect = raiser
            assert 500 == cli.get("/").status_code


def test_fallback_to_memory_config(redis_connection, extension_factory):
    _, limiter = extension_factory(
        config={ConfigVars.ENABLED: True},
        default_limits=["5/minute"],
        storage_uri="redis://localhost:46379",
        in_memory_fallback=["1/minute"],
    )
    assert len(limiter._in_memory_fallback) == 1
    assert limiter._in_memory_fallback_enabled

    _, limiter = extension_factory(
        config={ConfigVars.ENABLED: True, ConfigVars.IN_MEMORY_FALLBACK: "1/minute"},
        default_limits=["5/minute"],
        storage_uri="redis://localhost:46379",
    )
    assert len(limiter._in_memory_fallback) == 1
    assert limiter._in_memory_fallback_enabled

    _, limiter = extension_factory(
        config={ConfigVars.ENABLED: True, ConfigVars.IN_MEMORY_FALLBACK_ENABLED: True},
        default_limits=["5/minute"],
        storage_uri="redis://localhost:46379",
    )
    assert limiter._in_memory_fallback_enabled

    _, limiter = extension_factory(
        config={ConfigVars.ENABLED: True},
        default_limits=["5/minute"],
        storage_uri="redis://localhost:46379",
        in_memory_fallback_enabled=True,
    )


def test_fallback_to_memory_backoff_check(redis_connection, extension_factory):
    app, limiter = extension_factory(
        config={ConfigVars.ENABLED: True},
        default_limits=["5/minute"],
        storage_uri="redis://localhost:46379",
        in_memory_fallback=["1/minute"],
    )

    @app.route("/t1")
    def t1():
        return "test"

    with app.test_client() as cli:

        def raiser(*a):
            raise Exception("redis dead")

        with hiro.Timeline() as timeline:
            with patch("limits.storage.RedisStorage.incr") as incr:
                with patch("limits.storage.RedisStorage.check") as check:
                    check.return_value = False
                    incr.side_effect = raiser
                    assert cli.get("/t1").status_code == 200
                    assert cli.get("/t1").status_code == 429
                    timeline.forward(1)
                    assert cli.get("/t1").status_code == 429
                    timeline.forward(2)
                    assert cli.get("/t1").status_code == 429
                    timeline.forward(4)
                    assert cli.get("/t1").status_code == 429
                    timeline.forward(8)
                    assert cli.get("/t1").status_code == 429
                    timeline.forward(16)
                    assert cli.get("/t1").status_code == 429
                    timeline.forward(32)
                    assert cli.get("/t1").status_code == 200
            # redis back to normal, but exponential backoff will only
            # result in it being marked after pow(2,0) seconds and next
            # check
            assert cli.get("/t1").status_code == 429
            timeline.forward(2)
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 429


def test_fallback_to_memory_with_global_override(redis_connection, extension_factory):
    app, limiter = extension_factory(
        config={ConfigVars.ENABLED: True},
        default_limits=["5/minute"],
        storage_uri="redis://localhost:46379",
        in_memory_fallback=["1/minute"],
    )

    @app.route("/t1")
    def t1():
        return "test"

    @app.route("/t2")
    @limiter.limit("3 per minute")
    def t2():
        return "test"

    with app.test_client() as cli:
        assert cli.get("/t1").status_code == 200
        assert cli.get("/t1").status_code == 200
        assert cli.get("/t1").status_code == 200
        assert cli.get("/t1").status_code == 200
        assert cli.get("/t1").status_code == 200
        assert cli.get("/t1").status_code == 429
        assert cli.get("/t2").status_code == 200
        assert cli.get("/t2").status_code == 200
        assert cli.get("/t2").status_code == 200
        assert cli.get("/t2").status_code == 429

        def raiser(*a):
            raise Exception("redis dead")

        with patch("limits.storage.RedisStorage.incr") as incr:
            with patch("limits.storage.RedisStorage.check") as check:
                check.return_value = False
                incr.side_effect = raiser
                assert cli.get("/t1").status_code == 200
                assert cli.get("/t1").status_code == 429
                assert cli.get("/t2").status_code == 200
                assert cli.get("/t2").status_code == 429
        # redis back to normal, go back to regular limits
        with hiro.Timeline() as timeline:
            timeline.forward(2)
            limiter._storage.storage.flushall()
            assert cli.get("/t2").status_code == 200
            assert cli.get("/t2").status_code == 200
            assert cli.get("/t2").status_code == 200
            assert cli.get("/t2").status_code == 429


def test_fallback_to_memory(extension_factory):
    app, limiter = extension_factory(
        config={ConfigVars.ENABLED: True},
        default_limits=["2/minute"],
        storage_uri="redis://localhost:46379",
        in_memory_fallback_enabled=True,
        headers_enabled=True,
    )

    @app.route("/t1")
    def t1():
        return "test"

    @app.route("/t2")
    @limiter.limit("1 per minute")
    def t2():
        return "test"

    with app.test_client() as cli:
        assert cli.get("/t1").status_code == 200
        assert cli.get("/t1").status_code == 200
        assert cli.get("/t1").status_code == 429
        assert cli.get("/t2").status_code == 200
        assert cli.get("/t2").status_code == 429

        def raiser(*a):
            raise Exception("redis dead")

        with patch("limits.storage.RedisStorage.incr") as hit:
            hit.side_effect = raiser
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 200
            assert cli.get("/t1").status_code == 429
            assert cli.get("/t2").status_code == 200
            assert cli.get("/t2").status_code == 429
        with hiro.Timeline() as timeline:
            timeline.forward(1)
            limiter._storage.storage.flushall()
            assert cli.get("/t2").status_code == 200
            assert cli.get("/t2").status_code == 429
        with patch("limits.storage.RedisStorage.get") as get:
            get.side_effect = raiser
            assert cli.get("/t1").status_code == 200
